--
-- libxlsxwriter
--
-- Copyright 2014-2015, John McNamara, jmcnamara@cpan.org. See LICENSE.txt.
--

-- namespace xlsxwriter

include misc.e
include dll.e
include machine.e

atom libxlsxwriter

if platform() = WIN32 then
  libxlsxwriter = open_dll( "libxlsxwriter.dll" )
elsif platform() = LINUX then
  libxlsxwriter = open_dll( "libxlsxwriter.so" )
else
  crash_message( "This platform is not supported" )
  ? 0/0
end if

constant
  C_UINT8   = C_UCHAR,
  C_UINT16  = C_USHORT,
  C_UINT32  = C_UINT,
  C_INT32   = C_INT

constant
  x_format_set_align                = define_c_proc( libxlsxwriter, "+format_set_align", {C_POINTER,C_UINT8} ),
  x_format_set_bg_color             = define_c_proc( libxlsxwriter, "+format_set_bg_color", {C_POINTER,C_INT32} ),
  x_format_set_bold                 = define_c_proc( libxlsxwriter, "+format_set_bold", {C_POINTER} ),
  x_format_set_border               = define_c_proc( libxlsxwriter, "+format_set_border", {C_POINTER,C_UINT8} ),
  x_format_set_border_color         = define_c_proc( libxlsxwriter, "+format_set_border_color", {C_POINTER,C_INT32} ),
  x_format_set_bottom               = define_c_proc( libxlsxwriter, "+format_set_bottom", {C_POINTER,C_UINT8} ),
  x_format_set_bottom_color         = define_c_proc( libxlsxwriter, "+format_set_bottom_color", {C_POINTER,C_INT32} ),
  x_format_set_diag_border          = define_c_proc( libxlsxwriter, "+format_set_diag_border", {C_POINTER,C_UINT8} ),
  x_format_set_diag_color           = define_c_proc( libxlsxwriter, "+format_set_diag_color", {C_POINTER,C_INT32} ),
  x_format_set_diag_type            = define_c_proc( libxlsxwriter, "+format_set_diag_type", {C_POINTER,C_UINT8} ),
  x_format_set_fg_color             = define_c_proc( libxlsxwriter, "+format_set_fg_color", {C_POINTER,C_INT32} ),
  x_format_set_font_charset         = define_c_proc( libxlsxwriter, "+format_set_font_charset", {C_POINTER,C_UINT8} ),
  x_format_set_font_color           = define_c_proc( libxlsxwriter, "+format_set_font_color", {C_POINTER,C_INT32} ),
  x_format_set_font_condense        = define_c_proc( libxlsxwriter, "+format_set_font_condense", {C_POINTER} ),
  x_format_set_font_extend          = define_c_proc( libxlsxwriter, "+format_set_font_extend", {C_POINTER} ),
  x_format_set_font_family          = define_c_proc( libxlsxwriter, "+format_set_font_family", {C_POINTER,C_UINT8} ),
  x_format_set_font_name            = define_c_proc( libxlsxwriter, "+format_set_font_name", {C_POINTER,C_POINTER} ),
  x_format_set_font_outline         = define_c_proc( libxlsxwriter, "+format_set_font_outline", {C_POINTER} ),
  x_format_set_font_scheme          = define_c_proc( libxlsxwriter, "+format_set_font_scheme", {C_POINTER,C_POINTER} ),
  x_format_set_font_script          = define_c_proc( libxlsxwriter, "+format_set_font_script", {C_POINTER,C_UINT8} ),
  x_format_set_font_shadow          = define_c_proc( libxlsxwriter, "+format_set_font_shadow", {C_POINTER} ),
  x_format_set_font_size            = define_c_proc( libxlsxwriter, "+format_set_font_size", {C_POINTER,C_UINT16} ),
  x_format_set_font_strikeout       = define_c_proc( libxlsxwriter, "+format_set_font_strikeout", {C_POINTER} ),
  x_format_set_hidden               = define_c_proc( libxlsxwriter, "+format_set_hidden", {C_POINTER} ),
  x_format_set_indent               = define_c_proc( libxlsxwriter, "+format_set_indent", {C_POINTER,C_CHAR} ),
  x_format_set_italic               = define_c_proc( libxlsxwriter, "+format_set_italic", {C_POINTER} ),
  x_format_set_left                 = define_c_proc( libxlsxwriter, "+format_set_left", {C_POINTER,C_UINT8} ),
  x_format_set_left_color           = define_c_proc( libxlsxwriter, "+format_set_left_color", {C_POINTER,C_INT32} ),
  x_format_set_num_format           = define_c_proc( libxlsxwriter, "+format_set_num_format", {C_POINTER,C_POINTER} ),
  x_format_set_num_format_index     = define_c_proc( libxlsxwriter, "+format_set_num_format_index", {C_POINTER,C_UINT8} ),
  x_format_set_pattern              = define_c_proc( libxlsxwriter, "+format_set_pattern", {C_POINTER,C_UINT8} ),
  x_format_set_reading_order        = define_c_proc( libxlsxwriter, "+format_set_reading_order", {C_POINTER,C_UINT8} ),
  x_format_set_right                = define_c_proc( libxlsxwriter, "+format_set_right", {C_POINTER,C_UINT8} ),
  x_format_set_right_color          = define_c_proc( libxlsxwriter, "+format_set_right_color", {C_POINTER,C_INT32} ),
  x_format_set_rotation             = define_c_proc( libxlsxwriter, "+format_set_rotation", {C_POINTER,C_SHORT} ),
  x_format_set_shrink               = define_c_proc( libxlsxwriter, "+format_set_shrink", {C_POINTER} ),
  x_format_set_text_wrap            = define_c_proc( libxlsxwriter, "+format_set_text_wrap", {C_POINTER} ),
  x_format_set_theme                = define_c_proc( libxlsxwriter, "+format_set_theme", {C_POINTER,C_UINT8} ),
  x_format_set_top                  = define_c_proc( libxlsxwriter, "+format_set_top", {C_POINTER,C_UINT8} ),
  x_format_set_top_color            = define_c_proc( libxlsxwriter, "+format_set_top_color", {C_POINTER,C_INT32} ),
  x_format_set_underline            = define_c_proc( libxlsxwriter, "+format_set_underline", {C_POINTER,C_UINT8} ),
  x_format_set_unlocked             = define_c_proc( libxlsxwriter, "+format_set_unlocked", {C_POINTER} ),
  x_lxw_get_col                     = define_c_func( libxlsxwriter, "+lxw_get_col", {C_POINTER}, C_UINT32 ),
  x_lxw_get_col_2                   = define_c_func( libxlsxwriter, "+lxw_get_col_2", {C_POINTER}, C_UINT32 ),
  x_lxw_get_row                     = define_c_func( libxlsxwriter, "+lxw_get_row", {C_POINTER}, C_UINT16 ),
  x_lxw_get_row_2                   = define_c_func( libxlsxwriter, "+lxw_get_row_2", {C_POINTER}, C_UINT16 ),
  x_lxw_quote_sheetname             = define_c_func( libxlsxwriter, "+lxw_quote_sheetname", {C_POINTER}, C_POINTER ),
  x_new_workbook                    = define_c_func( libxlsxwriter, "+new_workbook", {C_POINTER}, C_POINTER ),
  x_new_workbook_opt                = define_c_func( libxlsxwriter, "+new_workbook_opt", {C_POINTER,C_POINTER}, C_POINTER ),
  x_workbook_add_format             = define_c_func( libxlsxwriter, "+workbook_add_format", {C_POINTER}, C_POINTER ),
  x_workbook_add_worksheet          = define_c_func( libxlsxwriter, "+workbook_add_worksheet", {C_POINTER,C_POINTER}, C_POINTER ),
  x_workbook_close                  = define_c_func( libxlsxwriter, "+workbook_close", {C_POINTER}, C_UINT8 ),
  x_workbook_define_name            = define_c_func( libxlsxwriter, "+workbook_define_name", {C_POINTER,C_POINTER, C_POINTER}, C_UINT8 ),
  x_worksheet_activate              = define_c_proc( libxlsxwriter, "+worksheet_activate", {C_POINTER} ),
  x_worksheet_autofilter            = define_c_func( libxlsxwriter, "+worksheet_autofilter", {C_POINTER,C_UINT32, C_UINT16,C_UINT32,C_UINT16}, C_CHAR ),
  x_worksheet_center_horizontally   = define_c_proc( libxlsxwriter, "+worksheet_center_horizontally", {C_POINTER} ),
  x_worksheet_center_vertically     = define_c_proc( libxlsxwriter, "+worksheet_center_vertically", {C_POINTER} ),
  x_worksheet_fit_to_pages          = define_c_proc( libxlsxwriter, "+worksheet_fit_to_pages", {C_POINTER,C_UINT16, C_UINT16} ),
  x_worksheet_gridlines             = define_c_proc( libxlsxwriter, "+worksheet_gridlines", {C_POINTER,C_UINT8} ),
  x_worksheet_merge_range           = define_c_func( libxlsxwriter, "+worksheet_merge_range", {C_POINTER,C_UINT32,C_UINT16,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_print_across          = define_c_proc( libxlsxwriter, "+worksheet_print_across", {C_POINTER} ),
  x_worksheet_print_area            = define_c_func( libxlsxwriter, "+worksheet_print_area", {C_POINTER,C_UINT32,C_UINT16,C_UINT32,C_UINT16}, C_UINT8 ),
  x_worksheet_print_row_col_headers = define_c_proc( libxlsxwriter, "+worksheet_print_row_col_headers", {C_POINTER} ),
  x_worksheet_repeat_columns        = define_c_func( libxlsxwriter, "+worksheet_repeat_columns", {C_POINTER,C_UINT16,C_UINT16}, C_UINT8 ),
  x_worksheet_repeat_rows           = define_c_func( libxlsxwriter, "+worksheet_repeat_rows", {C_POINTER,C_UINT32,C_UINT32}, C_UINT8 ),
  x_worksheet_select                = define_c_proc( libxlsxwriter, "+worksheet_select", {C_POINTER} ),
  x_worksheet_set_column            = define_c_func( libxlsxwriter, "+worksheet_set_column", {C_POINTER,C_UINT16,C_UINT16,C_DOUBLE,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_set_footer            = define_c_func( libxlsxwriter, "+worksheet_set_footer", {C_POINTER,C_POINTER}, C_UINT8 ),
  x_worksheet_set_footer_opt        = define_c_func( libxlsxwriter, "+worksheet_set_footer_opt", {C_POINTER,C_POINTER}, C_UINT8 ),
  x_worksheet_set_h_pagebreaks      = define_c_proc( libxlsxwriter, "+worksheet_set_h_pagebreaks", {C_POINTER,C_POINTER} ),
  x_worksheet_set_header            = define_c_func( libxlsxwriter, "+worksheet_set_header", {C_POINTER,C_POINTER}, C_UINT8 ),
  x_worksheet_set_header_opt        = define_c_func( libxlsxwriter, "+worksheet_set_header_opt", {C_POINTER,C_POINTER}, C_UINT8 ),
  x_worksheet_set_landscape         = define_c_proc( libxlsxwriter, "+worksheet_set_landscape", {C_POINTER} ),
  x_worksheet_set_margins           = define_c_proc( libxlsxwriter, "+worksheet_set_margins", {C_POINTER,C_DOUBLE,C_DOUBLE,C_DOUBLE,C_DOUBLE} ),
  x_worksheet_set_page_view         = define_c_proc( libxlsxwriter, "+worksheet_set_page_view", {C_POINTER} ),
  x_worksheet_set_paper             = define_c_proc( libxlsxwriter, "+worksheet_set_paper", {C_POINTER,C_UINT8} ),
  x_worksheet_set_portrait          = define_c_proc( libxlsxwriter, "+worksheet_set_portrait", {C_POINTER} ),
  x_worksheet_set_print_scale       = define_c_proc( libxlsxwriter, "+worksheet_set_print_scale", {C_POINTER,C_UINT16} ),
  x_worksheet_set_row               = define_c_func( libxlsxwriter, "+worksheet_set_row", {C_POINTER,C_UINT32,C_DOUBLE,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_set_start_page        = define_c_proc( libxlsxwriter, "+worksheet_set_start_page", {C_POINTER,C_UINT16} ),
  x_worksheet_set_v_pagebreaks      = define_c_proc( libxlsxwriter, "+worksheet_set_v_pagebreaks", {C_POINTER,C_POINTER} ),
  x_worksheet_write_array_formula   = define_c_func( libxlsxwriter, "+worksheet_write_array_formula", {C_POINTER,C_UINT32,C_UINT16,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_write_blank           = define_c_func( libxlsxwriter, "+worksheet_write_blank", {C_POINTER,C_UINT32,C_UINT16,C_POINTER}, C_CHAR ),
  x_worksheet_write_datetime        = define_c_func( libxlsxwriter, "+worksheet_write_datetime", {C_POINTER,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_write_formula         = define_c_func( libxlsxwriter, "+worksheet_write_formula", {C_POINTER,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_write_number          = define_c_func( libxlsxwriter, "+worksheet_write_number", {C_POINTER,C_UINT32,C_UINT16,C_DOUBLE,C_POINTER}, C_CHAR ),
  x_worksheet_write_string          = define_c_func( libxlsxwriter, "+worksheet_write_string", {C_POINTER,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_write_url             = define_c_func( libxlsxwriter, "+worksheet_write_url", {C_POINTER,C_UINT32,C_UINT16,C_POINTER,C_POINTER}, C_CHAR ),
  x_worksheet_write_url_opt         = define_c_func( libxlsxwriter, "+worksheet_write_url_opt", {C_POINTER,C_UINT32,C_UINT16,C_POINTER,C_POINTER,C_POINTER,C_POINTER}, C_CHAR )

global constant
  LXW_ROW_MAX            = 1048576,
  LXW_COL_MAX            = 16384,
  LXW_COL_META_MAX       = 128,
  LXW_HEADER_FOOTER_MAX  = 255

-- The Excel 2007 specification says that the maximum number of page
-- breaks is 1026. However, in practice it is actually 1023.
global constant LXW_BREAKS_MAX = 1023

-- Default column width in Excel
global constant LXW_DEF_COL_WIDTH = 8.43

-- Default row height in Excel
global constant LXW_DEF_ROW_HEIGHT = 15

-- Errors conditions encountered when closing the Workbook and writing the Excel file to disk.
global constant
  LXW_CLOSE_ERROR_NONE = 0,  -- No error
  LXW_CLOSE_ERROR_ZIP  = 1  -- Error encountered when creating file zip container

global type lxw_close_error(integer x)
  return find(x, {LXW_CLOSE_ERROR_NONE, LXW_CLOSE_ERROR_ZIP})
end type

global constant
  stqh_first         = 0,  -- first element
  stqh_last          = 4,  -- addr of last next element
  SIZEOF_STAILQ_HEAD = 8

global constant
  stqe_next           = 0,  -- next element
  SIZEOF_STAILQ_ENTRY = 4

global constant
  lxw_datetime__year  =  0, -- int
  lxw_datetime__month =  4, -- int
  lxw_datetime__day   =  8, -- int
  lxw_datetime__hour  = 12, -- int
  lxw_datetime__min   = 16, -- int
  lxw_datetime__sec   = 20, -- double
  SIZEOF_LXW_DATETIME = 32

global constant
  lxw_workbook_options__constant_memory = 0, -- uint8_t
  SIZEOF_LXW_WORKBOOK_OPTIONS            = 1

global constant
  lxw_workbook__file             =   0, -- FILE*
  lxw_workbook__worksheets       =   4, -- lxw_worksheets*
  lxw_workbook__formats          =   8, -- lxw_formats*
  lxw_workbook__defined_names    =  12, -- lxw_defined_names*
  lxw_workbook__sst              =  16, -- lxw_sst*
  lxw_workbook__properties       =  20, -- lxw_properties*
  lxw_workbook__filename         =  24, -- const char*
  lxw_workbook__options          =  28, -- lxw_workbook_options*
  lxw_workbook__num_sheets       =  30, -- uint16_t
  lxw_workbook__first_sheet      =  32, -- uint16_t
  lxw_workbook__active_sheet     =  36, -- uint32_t
  lxw_workbook__num_xf_formats   =  40, -- uint16_t
  lxw_workbook__num_format_count =  42, -- uint16_t
  lxw_workbook__font_count       =  44, -- uint16_t
  lxw_workbook__border_count     =  46, -- uint16_t
  lxw_workbook__fill_count       =  48, -- uint16_t
  lxw_workbook__optimize         =  50, -- uint8_t
  lxw_workbook__used_xf_formats  =  52, -- lxw_hash_table*
  SIZEOF_LXW_WORKBOOK             =  56

global constant
  lxw_worksheet__file                  =   0, -- FILE *
  lxw_worksheet__optimize_tmpfile      =   4, -- FILE *
  lxw_worksheet__table                 =   8, -- lxw_table_rows *
  lxw_worksheet__hyperlinks            =  12, -- lxw_table_rows *
  lxw_worksheet__array                 =  16, -- lxw_cell**
  lxw_worksheet__merged_ranges         =  20, -- lxw_merged_ranges*
  lxw_worksheet__dim_rowmin            =  24, -- lxw_row_t
  lxw_worksheet__dim_rowmax            =  28, -- lxw_row_t
  lxw_worksheet__dim_colmin            =  32, -- lxw_col_t
  lxw_worksheet__dim_colmax            =  34, -- lxw_col_t
  lxw_worksheet__sst                   =  36, -- lxw_sst*
  lxw_worksheet__name                  =  40, -- char*
  lxw_worksheet__quoted_name           =  44, -- char*
  lxw_worksheet__index                 =  48, -- uint32_t
  lxw_worksheet__active                =  52, -- uint8_t
  lxw_worksheet__selected              =  53, -- uint8_t
  lxw_worksheet__hidden                =  54, -- uint8_t
  lxw_worksheet__active_sheet          =  56, -- uint32_t
  lxw_worksheet__col_options           =  60, -- lxw_col_options**
  lxw_worksheet__col_options_max       =  64, -- uint16_t
  lxw_worksheet__col_sizes             =  68, -- double*
  lxw_worksheet__col_sizes_max         =  72, -- uint16_t
  lxw_worksheet__col_formats           =  76, -- lxw_format**
  lxw_worksheet__col_formats_max       =  80, -- uint16_t
  lxw_worksheet__col_size_changed      =  82, -- uint8_t
  lxw_worksheet__optimize              =  83, -- uint8_t
  lxw_worksheet__optimize_row          =  84, -- lxw_row*
  lxw_worksheet__fit_width             =  90, -- uint16_t
  lxw_worksheet__fit_height            =  88, -- uint16_t
  lxw_worksheet__horizontal_dpi        =  92, -- uint16_t
  lxw_worksheet__hlink_count           =  94, -- uint16_t
  lxw_worksheet__page_start            =  96, -- uint16_t
  lxw_worksheet__print_scale           =  98, -- uint16_t
  lxw_worksheet__rel_count             = 100, -- uint16_t
  lxw_worksheet__vertical_dpi          = 102, -- uint16_t
  lxw_worksheet__filter_on             = 104, -- uint8_t
  lxw_worksheet__fit_page              = 105, -- uint8_t
  lxw_worksheet__hcenter               = 106, -- uint8_t
  lxw_worksheet__orientation           = 107, -- uint8_t
  lxw_worksheet__outline_changed       = 108, -- uint8_t
  lxw_worksheet__page_order            = 109, -- uint8_t
  lxw_worksheet__page_setup_changed    = 110, -- uint8_t
  lxw_worksheet__page_view             = 111, -- uint8_t
  lxw_worksheet__paper_size            = 112, -- uint8_t
  lxw_worksheet__print_gridlines       = 113, -- uint8_t
  lxw_worksheet__print_headers         = 114, -- uint8_t
  lxw_worksheet__print_options_changed = 115, -- uint8_t
  lxw_worksheet__screen_gridlines      = 116, -- uint8_t
  lxw_worksheet__tab_color             = 117, -- uint8_t
  lxw_worksheet__vba_codename          = 118, -- uint8_t
  lxw_worksheet__vcenter               = 119, -- uint8_t
  lxw_worksheet__margin_left           = 120, -- double
  lxw_worksheet__margin_right          = 128, -- double
  lxw_worksheet__margin_top            = 136, -- double
  lxw_worksheet__margin_bottom         = 144, -- double
  lxw_worksheet__margin_header         = 152, -- double
  lxw_worksheet__margin_footer         = 160, -- double
  lxw_worksheet__header_footer_changed = 168, -- uint8_t
  lxw_worksheet__header                = 169, -- char[LXW_HEADER_FOOTER_MAX]
  lxw_worksheet__footer                = 424, -- char[LXW_HEADER_FOOTER_MAX]
  lxw_worksheet__repeat_rows           = 680, -- lxw_repeat_rows
  lxw_worksheet__repeat_cols           = 692, -- lxw_repeat_cols
  lxw_worksheet__print_area            = 700, -- lxw_print_area
  lxw_worksheet__autofilter            = 716, -- lxw_autofilter
  lxw_worksheet__merged_range_count    = 732, -- uint16_t
  lxw_worksheet__hbreaks               = 736, -- lxw_row_t*
  lxw_worksheet__vbreaks               = 740, -- lxw_col_t*
  lxw_worksheet__external_hyperlinks   = 744, -- lxw_rel_tuples*
  lxw_worksheet__list_pointers         = 748, -- lxw_worksheet*
  SIZEOF_LXW_WORKSHEET                  = 752

global constant
  lxw_worksheets__stqh_first = 0, -- pointer
  lxw_worksheets__stqh_last  = 4, -- pointer
  SIZEOF_LXW_WORKSHEETS      = 8

-- Error codes from `worksheet_write*()` functions.

global constant
  LXW_WRITE_ERROR_NONE    = 0,  -- No error.
  LXW_RANGE_ERROR         = 1,  -- Row or column index out of range.
  LXW_STRING_LENGTH_ERROR = 2,  -- String exceeds Excel's LXW_STRING_LENGTH_ERROR limit.
  LXW_STRING_HASH_ERROR   = 3   -- Error finding string index.

global type lxw_write_error(integer x)
  return find(x, {LXW_WRITE_ERROR_NONE, LXW_RANGE_ERROR, LXW_STRING_LENGTH_ERROR,
                  LXW_STRING_HASH_ERROR})
end type

-- Gridline options using in `worksheet_gridlines()`.

global constant
  LXW_HIDE_ALL_GRIDLINES    = 0,  -- Hide screen and print gridlines.
  LXW_SHOW_SCREEN_GRIDLINES = 1,  -- Show screen gridlines.
  LXW_SHOW_PRINT_GRIDLINES  = 2,  -- Show print gridlines.
  LXW_SHOW_ALL_GRIDLINES    = 3   -- Show screen and print gridlines.

global type lxw_gridlines(integer x)
  return find(x, {LXW_HIDE_ALL_GRIDLINES, LXW_SHOW_SCREEN_GRIDLINES,
                  LXW_SHOW_PRINT_GRIDLINES, LXW_SHOW_ALL_GRIDLINES})
end type

global constant
  LXW_FORMAT_FIELD_LEN    = 128,
  LXW_DEFAULT_FONT_NAME   = "Calibri",
  LXW_DEFAULT_FONT_FAMILY = 2,
  LXW_DEFAULT_FONT_THEME  = 1,
  LXW_PROPERTY_UNSET      = -1,
  LXW_COLOR_UNSET         = -1,
  LXW_COLOR_MASK          = #FFFFFF,
  LXW_MIN_FONT_SIZE       = 1,
  LXW_MAX_FONT_SIZE       = 409

-- Format underline values for format_set_underline().
global constant
  LXW_UNDERLINE_SINGLE            = 1,  -- Single underline
  LXW_UNDERLINE_DOUBLE            = 2,  -- Double underline
  LXW_UNDERLINE_SINGLE_ACCOUNTING = 3,  -- Single accounting underline
  LXW_UNDERLINE_DOUBLE_ACCOUNTING = 4   -- Double accounting underline

global type lxw_format_underlines(integer x)
  return find(x, {LXW_UNDERLINE_SINGLE, LXW_UNDERLINE_DOUBLE,
                  LXW_UNDERLINE_SINGLE_ACCOUNTING, LXW_UNDERLINE_DOUBLE_ACCOUNTING})
end type

-- Superscript and subscript values for format_set_font_script().
global constant
  LXW_FONT_SUPERSCRIPT = 1,  -- Superscript font
  LXW_FONT_SUBSCRIPT   = 2   -- Subscript font

global type lxw_format_scripts(integer x)
  return find(x, {LXW_FONT_SUPERSCRIPT, LXW_FONT_SUBSCRIPT})
end type

-- Alignment values for format_set_align().
global constant
  LXW_ALIGN_NONE                 =  0,  -- No alignment. Cell will use Excel's default for the data type
  LXW_ALIGN_LEFT                 =  1,  -- Left horizontal alignment
  LXW_ALIGN_CENTER               =  2,  -- Center horizontal alignment
  LXW_ALIGN_RIGHT                =  3,  -- Right horizontal alignment
  LXW_ALIGN_FILL                 =  4,  -- Cell fill horizontal alignment
  LXW_ALIGN_JUSTIFY              =  5,  -- Justify horizontal alignment
  LXW_ALIGN_CENTER_ACROSS        =  6,  -- Center Across horizontal alignment
  LXW_ALIGN_DISTRIBUTED          =  7,  -- Left horizontal alignment
  LXW_ALIGN_VERTICAL_TOP         =  8,  -- Top vertical alignment
  LXW_ALIGN_VERTICAL_BOTTOM      =  9,  -- Bottom vertical alignment
  LXW_ALIGN_VERTICAL_CENTER      = 10,  -- Center vertical alignment
  LXW_ALIGN_VERTICAL_JUSTIFY     = 11,  -- Justify vertical alignment
  LXW_ALIGN_VERTICAL_DISTRIBUTED = 12  -- Distributed vertical alignment

global type lxw_format_alignments(integer x)
  return find(x, {LXW_ALIGN_NONE, LXW_ALIGN_LEFT, LXW_ALIGN_CENTER,
                  LXW_ALIGN_RIGHT, LXW_ALIGN_FILL, LXW_ALIGN_JUSTIFY,
                  LXW_ALIGN_CENTER_ACROSS, LXW_ALIGN_DISTRIBUTED,
                  LXW_ALIGN_VERTICAL_TOP, LXW_ALIGN_VERTICAL_BOTTOM,
                  LXW_ALIGN_VERTICAL_CENTER, LXW_ALIGN_VERTICAL_JUSTIFY,
                  LXW_ALIGN_VERTICAL_DISTRIBUTED})
end type

global constant
  LXW_DIAGONAL_BORDER_UP      = 1,
  LXW_DIAGONAL_BORDER_DOWN    = 2,
  LXW_DIAGONAL_BORDER_UP_DOWN = 3

global type lxw_format_diagonal_types(integer x)
  return find(x, {LXW_DIAGONAL_BORDER_UP, LXW_DIAGONAL_BORDER_DOWN,
                  LXW_DIAGONAL_BORDER_UP_DOWN})
end type

-- Predefined values for common colors.
global constant
  LXW_COLOR_BLACK   = #000000,  -- Black
  LXW_COLOR_BLUE    = #0000FF,  -- Blue
  LXW_COLOR_BROWN   = #800000,  -- Brown
  LXW_COLOR_CYAN    = #00FFFF,  -- Cyan
  LXW_COLOR_GRAY    = #808080,  -- Gray
  LXW_COLOR_GREEN   = #008000,  -- Green
  LXW_COLOR_LIME    = #00FF00,  -- Lime
  LXW_COLOR_MAGENTA = #FF00FF,  -- Magenta
  LXW_COLOR_NAVY    = #000080,  -- Navy
  LXW_COLOR_ORANGE  = #FF6600,  -- Orange
  LXW_COLOR_PINK    = #FF00FF,  -- Pink
  LXW_COLOR_PURPLE  = #800080,  -- Purple
  LXW_COLOR_RED     = #FF0000,  -- Red
  LXW_COLOR_SILVER  = #C0C0C0,  -- Silver
  LXW_COLOR_WHITE   = #FFFFFF,  -- White
  LXW_COLOR_YELLOW  = #FFFF00   -- Yellow

global type lxw_defined_colors(integer x)
  return find(x, {LXW_COLOR_BLACK, LXW_COLOR_BLUE, LXW_COLOR_BROWN,
                  LXW_COLOR_CYAN, LXW_COLOR_GRAY, LXW_COLOR_GREEN,
                  LXW_COLOR_LIME, LXW_COLOR_MAGENTA, LXW_COLOR_NAVY,
                  LXW_COLOR_ORANGE, LXW_COLOR_PINK, LXW_COLOR_PURPLE,
                  LXW_COLOR_RED, LXW_COLOR_SILVER, LXW_COLOR_WHITE,
                  LXW_COLOR_YELLOW})
end type

-- Pattern value for use with format_set_pattern().
global constant
  LXW_PATTERN_NONE             =  0,  -- Empty pattern
  LXW_PATTERN_SOLID            =  1,  -- Solid pattern
  LXW_PATTERN_MEDIUM_GRAY      =  2,  -- Medium gray pattern
  LXW_PATTERN_DARK_GRAY        =  3,  -- Dark gray pattern
  LXW_PATTERN_LIGHT_GRAY       =  4,  -- Light gray pattern
  LXW_PATTERN_DARK_HORIZONTAL  =  5,  -- Dark horizontal line pattern
  LXW_PATTERN_DARK_VERTICAL    =  6,  -- Dark vertical line pattern
  LXW_PATTERN_DARK_DOWN        =  7,  -- Dark diagonal stripe pattern
  LXW_PATTERN_DARK_UP          =  8,  -- Reverse dark diagonal stripe pattern
  LXW_PATTERN_DARK_GRID        =  9,  -- Dark grid pattern
  LXW_PATTERN_DARK_TRELLIS     = 10,  -- Dark trellis pattern
  LXW_PATTERN_LIGHT_HORIZONTAL = 11,  -- Light horizontal Line pattern
  LXW_PATTERN_LIGHT_VERTICAL   = 12,  -- Light vertical line pattern
  LXW_PATTERN_LIGHT_DOWN       = 13,  -- Light diagonal stripe pattern
  LXW_PATTERN_LIGHT_UP         = 14,  -- Reverse light diagonal stripe pattern
  LXW_PATTERN_LIGHT_GRID       = 15,  -- Light grid pattern
  LXW_PATTERN_LIGHT_TRELLIS    = 16,  -- Light trellis pattern
  LXW_PATTERN_GRAY_125         = 17,  -- 12.5% gray pattern
  LXW_PATTERN_GRAY_0625        = 18  -- 6.25% gray pattern

global type lxw_format_patterns(integer x)
  return find(x, {LXW_PATTERN_NONE, LXW_PATTERN_SOLID,
                  LXW_PATTERN_MEDIUM_GRAY, LXW_PATTERN_DARK_GRAY,
                  LXW_PATTERN_LIGHT_GRAY, LXW_PATTERN_DARK_HORIZONTAL,
                  LXW_PATTERN_DARK_VERTICAL, LXW_PATTERN_DARK_DOWN,
                  LXW_PATTERN_DARK_UP, LXW_PATTERN_DARK_GRID,
                  LXW_PATTERN_DARK_TRELLIS, LXW_PATTERN_LIGHT_HORIZONTAL,
                  LXW_PATTERN_LIGHT_VERTICAL, LXW_PATTERN_LIGHT_DOWN,
                  LXW_PATTERN_LIGHT_UP, LXW_PATTERN_LIGHT_GRID,
                  LXW_PATTERN_LIGHT_TRELLIS, LXW_PATTERN_GRAY_125,
                  LXW_PATTERN_GRAY_0625})
end type

-- Cell border styles for use with format_set_border().
global constant
  LXW_BORDER_NONE                =  1,  -- No border
  LXW_BORDER_THIN                =  2,  -- Thin border style
  LXW_BORDER_MEDIUM              =  3,  -- Medium border style
  LXW_BORDER_DASHED              =  4,  -- Dashed border style
  LXW_BORDER_DOTTED              =  5,  -- Dotted border style
  LXW_BORDER_THICK               =  6,  -- Thick border style
  LXW_BORDER_DOUBLE              =  7,  -- Double border style
  LXW_BORDER_HAIR                =  8,  -- Hair border style
  LXW_BORDER_MEDIUM_DASHED       =  9,  -- Medium dashed border style
  LXW_BORDER_DASH_DOT            = 10,  -- Dash-dot border style
  LXW_BORDER_MEDIUM_DASH_DOT     = 11,  -- Medium dash-dot border style
  LXW_BORDER_DASH_DOT_DOT        = 12,  -- Dash-dot-dot border style
  LXW_BORDER_MEDIUM_DASH_DOT_DOT = 13,  -- Medium dash-dot-dot border style
  LXW_BORDER_SLANT_DASH_DOT      = 14   -- Slant dash-dot border style

global type lxw_format_borders(integer x)
  return find(x, {LXW_BORDER_NONE, LXW_BORDER_THIN, LXW_BORDER_MEDIUM,
                  LXW_BORDER_DASHED, LXW_BORDER_DOTTED, LXW_BORDER_THICK,
                  LXW_BORDER_DOUBLE, LXW_BORDER_HAIR, LXW_BORDER_MEDIUM_DASHED,
                  LXW_BORDER_DASH_DOT, LXW_BORDER_MEDIUM_DASH_DOT, LXW_BORDER_DASH_DOT_DOT,
                  LXW_BORDER_MEDIUM_DASH_DOT_DOT, LXW_BORDER_SLANT_DASH_DOT})
end type

-----------------------------------------------------------------------------

-- next routines try to partially replace map

function map_find(sequence map, object key)
  for i = 1 to length(map) do
    if equal(map[i][1], key) then  -- key matches
      return i
    end if
  end for
  return 0
end function

function map_get(sequence map, object key, object default)
  integer i

  i = map_find(map, key)
  if i then  -- key matches
    return map[i][2]
  else
    return default
  end if
end function

function map_set(sequence map, object key, object value, integer add)
  integer i

  i = map_find(map, key)
  if i then  -- key matches
    map[i][2] = value
  else
    if add then
      map = append(map, {key, value})
    end if
  end if
  return map
end function

function map_remove(sequence map, object key)
  sequence result
  
  result = {}
  for i = 1 to length(map) do
    if not equal(map[i][1], key) then
      result = append(result, map[i])
    end if
  end for
  return result
end function

-----------------------------------------------------------------------------

-- next routines help manage memory allocation

function allocateString(object sheetname)
  if sequence( sheetname ) then
    return allocate_string( sheetname )
  end if
  return sheetname
end function

procedure freeMemory(atom sheetname)
  if sheetname then free( sheetname ) end if
end procedure

-----------------------------------------------------------------------------

function peek_ptr( atom addr )
  return peek4u( addr )
end function

procedure poke_int( atom addr, atom value )
  poke4( addr, value )
end procedure

procedure poke_dbl( atom addr, atom value )
  poke( addr, atom_to_float64(value) )
end procedure

-- Allocate a datetime structure from a Euphoria datetime value.
--
-- 'datetime' should be {year, month, day, hour, minute, second}

global function allocate_datetime( sequence datetime )
  atom ptr
  integer lg

  ptr = allocate( SIZEOF_LXW_DATETIME )
  mem_set( ptr, NULL, SIZEOF_LXW_DATETIME )
  -- not-so-easy!
  lg = length( datetime )
  if lg = 6 then poke_dbl( ptr + lxw_datetime__sec,   datetime[6] )
  elsif lg = 5 then poke_int( ptr + lxw_datetime__min,   datetime[5] )
  elsif lg = 4 then poke_int( ptr + lxw_datetime__hour,   datetime[4] )
  elsif lg = 3 then poke_int( ptr + lxw_datetime__day,   datetime[3] )
  elsif lg = 2 then poke_int( ptr + lxw_datetime__month, datetime[2] )
  elsif lg = 1 then poke_int( ptr + lxw_datetime__year,   datetime[1] )
  end if
  return ptr
end function

-- Allocate a NULL terminated array of 4-byte values.

global function allocate_array4( sequence data )
  atom size, ptr

  size = 4 * (length(data)+1)
  ptr = allocate( size )
  poke4( ptr, data & {NULL} )
  return ptr
end function

function peek_string(atom a)
  integer i
  sequence s

  s = ""
  if a then
    i = peek(a)
    while i do
      s = append(s, i)
      a += 1
      i = peek(a)
    end while
  end if
  return s
end function

-- N.B. We have to track the filename object in memory for the lifetime of each Workbook.
sequence m_filename
m_filename = {}

-- Create a new workbook object with (optional) additional workbook options.

-- options = NULL by default
global function new_workbook( object filename, atom options )
  atom result

  result = NULL
  if atom( filename ) then filename = peek_string( filename ) end if
  filename = allocateString( filename )
  if options = NULL then
    result = c_func( x_new_workbook, {filename} )
  else
    result = c_func( x_new_workbook_opt, {filename,options} )
  end if
  m_filename = map_set(m_filename, result, filename, 1)
  return result
end function

-- Create a new workbook object with additional workbook options.

global function new_workbook_opt( object filename, atom options )
  atom result

  if atom( filename ) then filename = peek_string( filename ) end if
  filename = allocateString( filename )
  result = c_func( x_new_workbook_opt, {filename,options} )
  m_filename = map_set(m_filename, result, filename, 1)
  return result
end function

-- Create a new Format object to format cells in worksheets.

global function workbook_add_format( atom workbook )
  return c_func( x_workbook_add_format, {workbook} )
end function

-- Add a new worksheet to a workbook with (optional) the provided name.

-- sheetname = NULL by default
global function workbook_add_worksheet( atom workbook, object sheetname )
  object result

  sheetname = allocateString( sheetname )
  result = c_func( x_workbook_add_worksheet, {workbook,sheetname} )
  freeMemory(sheetname)
  return result
end function

-- Close the Workbook object and write the XLSX file.

global function workbook_close( atom workbook )
  atom result, filename
  
  result = c_func( x_workbook_close, {workbook} )
  if result = 0 then
    -- clean up the workbook filename object
    filename = map_get( m_filename, workbook, NULL )
    if filename != NULL then free( filename ) end if
    m_filename = map_remove( m_filename, workbook )
  end if
  return result
end function

-- Return a list of all Worksheets in the Workbook.

global function workbook_get_worksheets( atom workbook )
  sequence result
  atom worksheets, next

  result = {}
  worksheets = peek_ptr( workbook + lxw_workbook__worksheets )
  next = peek_ptr( worksheets + stqh_first )
  while next != NULL do
    result = append( result, next )
    next = peek_ptr( next + lxw_worksheet__list_pointers + stqe_next )
  end while
  return result
end function

-- Create a defined name in the workbook to use as a variable.

global function workbook_define_name( atom workbook, object name, object formula )
  object result

  name = allocateString( name )
  formula = allocateString( formula )
  result = c_func( x_workbook_define_name, {workbook,name,formula} )
  freeMemory(name)
  freeMemory(formula)
  return result
end function

-- Write a number to a worksheet cell.

-- format = NULL by default
global function worksheet_write_number( atom worksheet, sequence cell, atom number, atom format )
  return c_func( x_worksheet_write_number, {worksheet,cell[1],cell[2],number,format} )
end function

-- Write a string to a worksheet cell.

-- format = NULL by default
global function worksheet_write_string( atom worksheet, sequence cell, object string, atom format )
  object result

  string = allocateString( string )
  result = c_func( x_worksheet_write_string, {worksheet,cell[1],cell[2],string,format} )
  freeMemory(string)
  return result
end function

-- Write a formula to a worksheet cell.

-- format = NULL by default
global function worksheet_write_formula( atom worksheet, sequence cell, object formula, atom format )
  object result

  formula = allocateString( formula )
  result = c_func( x_worksheet_write_formula, {worksheet,cell[1],cell[2],formula,format} )
  freeMemory(formula)
  return result
end function

-- Write an array formula to a worksheet cell.

-- format = NULL by default
global function worksheet_write_array_formula( atom worksheet, sequence range, object formula, atom format )
  object result

  formula = allocateString( formula )
  result = c_func( x_worksheet_write_array_formula, {worksheet,range[1],range[2],range[3],range[4],formula,format} )
  freeMemory(formula)
  return result
end function

-- Write a date or time to a worksheet cell.

-- format = NULL by default
global function worksheet_write_datetime( atom worksheet, sequence cell, object datetime, atom format )
  object result

  if sequence(datetime) then datetime = allocate_datetime( datetime ) end if
  result = c_func( x_worksheet_write_datetime, {worksheet,cell[1],cell[2],datetime,format} )
  freeMemory(datetime)
  return result
end function

-- Write a URL/hyperlink to a worksheet cell.

-- string, tooltip and format = NULL by default
global function worksheet_write_url( atom worksheet, sequence cell, object url, object string,
    object tooltip, atom format )
  object result, str

  url = allocateString( url )
  str = allocateString( string )
  tooltip = allocateString( tooltip )

  if string = NULL and tooltip = NULL then
    result = c_func( x_worksheet_write_url, {worksheet,cell[1],cell[2],url,format} )
  else
    result = c_func( x_worksheet_write_url_opt, {worksheet,cell[1],cell[2],url,str,tooltip,format} )
  end if

  freeMemory(url)
  if sequence(string) then freeMemory(str) end if
  freeMemory(tooltip)
  return result
end function

-- Write a formatted blank worksheet cell.

-- format = NULL by default
global function worksheet_write_blank( atom worksheet, sequence cell, atom format )
  return c_func( x_worksheet_write_blank, {worksheet,cell[1],cell[2],format} )
end function

-- Set the properties for a row of cells.

-- format and options = NULL by default
global function worksheet_set_row( atom worksheet, atom row, atom height, atom format, atom options )
  return c_func( x_worksheet_set_row, {worksheet,row,height,format,options} )
end function

-- Set the properties for one or more columns of cells.

-- format and options = NULL by default
global function worksheet_set_column( atom worksheet, sequence cols, atom width, atom format, atom options )
  return c_func( x_worksheet_set_column, {worksheet,cols[1],cols[2],width,format,options} )
end function

-- Merge a range of cells.

-- format = NULL by default
global function worksheet_merge_range( atom worksheet, sequence range, object string, atom format )
  object result

  string = allocateString( string )
  result = c_func( x_worksheet_merge_range, {worksheet,range[1],range[2],range[3],range[4],string,format} )
  freeMemory(string)
  return result
end function

-- Set the autofilter area in the worksheet.

global function worksheet_autofilter( atom worksheet, sequence range )
  return c_func( x_worksheet_autofilter, {worksheet,range[1],range[2],range[3],range[4]} )
end function

-- Make a worksheet the active, i.e., visible worksheet.

global procedure worksheet_activate( atom worksheet )
  c_proc( x_worksheet_activate, {worksheet} )
end procedure

-- Set a worksheet tab as selected.

global procedure worksheet_select( atom worksheet )
  c_proc( x_worksheet_select, {worksheet} )
end procedure

-- Set the page orientation as landscape.

global procedure worksheet_set_landscape( atom worksheet )
  c_proc( x_worksheet_set_landscape, {worksheet} )
end procedure

-- Set the page orientation as portrait.

global procedure worksheet_set_portrait( atom worksheet )
  c_proc( x_worksheet_set_portrait, {worksheet} )
end procedure

-- Set the page layout to page view mode.

global procedure worksheet_set_page_view( atom worksheet )
  c_proc( x_worksheet_set_page_view, {worksheet} )
end procedure

-- Set the paper type for printing.
--
--   | Index | Paper format         | Paper size        |
--   | ----- | -------------------- | ----------------- |
--   |   0   | Printer default      | Printer default   |
--   |   1   | Letter               | 8 1/2 x 11 in     |
--   |   2   | Letter Small         | 8 1/2 x 11 in     |
--   |   3   | Tabloid              | 11 x 17 in        |
--   |   4   | Ledger               | 17 x 11 in        |
--   |   5   | Legal                | 8 1/2 x 14 in     |
--   |   6   | Statement            | 5 1/2 x 8 1/2 in  |
--   |   7   | Executive            | 7 1/4 x 10 1/2 in |
--   |   8   | A3                   | 297 x 420 mm      |
--   |   9   | A4                   | 210 x 297 mm      |
--   |  10   | A4 Small             | 210 x 297 mm      |
--   |  11   | A5                   | 148 x 210 mm      |
--   |  12   | B4                   | 250 x 354 mm      |
--   |  13   | B5                   | 182 x 257 mm      |
--   |  14   | Folio                | 8 1/2 x 13 in     |
--   |  15   | Quarto               | 215 x 275 mm      |
--   |  16   | ---                  | 1#14 in           |
--   |  17   | ---                  | 11x17 in          |
--   |  18   | Note                 | 8 1/2 x 11 in     |
--   |  19   | Envelope 9           | 3 7/8 x 8 7/8     |
--   |  20   | Envelope 10          | 4 1/8 x 9 1/2     |
--   |  21   | Envelope 11          | 4 1/2 x 10 3/8    |
--   |  22   | Envelope 12          | 4 3/4 x 11        |
--   |  23   | Envelope 14          | 5 x 11 1/2        |
--   |  24   | C size sheet         | ---               |
--   |  25   | D size sheet         | ---               |
--   |  26   | E size sheet         | ---               |
--   |  27   | Envelope DL          | 110 x 220 mm      |
--   |  28   | Envelope C3          | 324 x 458 mm      |
--   |  29   | Envelope C4          | 229 x 324 mm      |
--   |  30   | Envelope C5          | 162 x 229 mm      |
--   |  31   | Envelope C6          | 114 x 162 mm      |
--   |  32   | Envelope C65         | 114 x 229 mm      |
--   |  33   | Envelope B4          | 250 x 353 mm      |
--   |  34   | Envelope B5          | 176 x 250 mm      |
--   |  35   | Envelope B6          | 176 x 125 mm      |
--   |  36   | Envelope             | 110 x 230 mm      |
--   |  37   | Monarch              | 3.875 x 7.5 in    |
--   |  38   | Envelope             | 3 5/8 x 6 1/2 in  |
--   |  39   | Fanfold              | 14 7/8 x 11 in    |
--   |  40   | German Std Fanfold   | 8 1/2 x 12 in     |
--   |  41   | German Legal Fanfold | 8 1/2 x 13 in     |
--

-- paper_type = 0 by default
global procedure worksheet_set_paper( atom worksheet, atom paper_type )
  c_proc( x_worksheet_set_paper, {worksheet,paper_type} )
end procedure

-- Set the worksheet margins (in inches) for the printed page.

-- left, right, top and bottom = -1 by default
global procedure worksheet_set_margins( atom worksheet, atom left, atom right, atom top, atom bottom )
  c_proc( x_worksheet_set_margins, {worksheet,left,right,top,bottom} )
end procedure

-- Set the printed page header caption with (optional) additional options.
--
--   | Control         | Category      | Description           |
--   | --------------- | ------------- | --------------------- |
--   | `&L`            | Justification | Left                  |
--   | `&C`            |               | Center                |
--   | `&R`            |               | Right                 |
--   | `&P`            | Information   | Page number           |
--   | `&N`            |               | Total number of pages |
--   | `&D`            |               | Date                  |
--   | `&T`            |               | Time                  |
--   | `&F`            |               | File name             |
--   | `&A`            |               | Worksheet name        |
--   | `&Z`            |               | Workbook path         |
--   | `&fontsize`     | Font          | Font size             |
--   | `&"font,style"` |               | Font name and style   |
--   | `&U`            |               | Single underline      |
--   | `&E`            |               | Double underline      |
--   | `&S`            |               | Strikethrough         |
--   | `&X`            |               | Superscript           |
--   | `&Y`            |               | Subscript             |

-- options = NULL by default
global function worksheet_set_header( atom worksheet, object string, atom options )
  object result

  string = allocateString( string )

  if options = NULL then
    result = c_func( x_worksheet_set_header, {worksheet,string} )
  else
    result = c_func( x_worksheet_set_header_opt, {worksheet,string,options} )
  end if

  freeMemory(string)
  return result
end function

-- Set the printed page footer caption with (optional) additional options.

-- options = NULL by default
global function worksheet_set_footer( atom worksheet, object string, atom options )
  object result

  string = allocateString( string )

  if options = NULL then
    result = c_func( x_worksheet_set_footer, {worksheet,string} )
  else
    result = c_func( x_worksheet_set_footer_opt, {worksheet,string,options} )
  end if

  freeMemory(string)
  return result
end function

-- Set the horizontal page breaks on a worksheet.

global procedure worksheet_set_h_pagebreaks( atom worksheet, object breaks )
  if sequence( breaks ) then breaks = allocate_array4( breaks ) end if
  c_proc( x_worksheet_set_h_pagebreaks, {worksheet,breaks} )
end procedure

-- Set the vertical page breaks on a worksheet.

global procedure worksheet_set_v_pagebreaks( atom worksheet, object breaks )
  if sequence( breaks ) then breaks = allocate_array4( breaks ) end if
  c_proc( x_worksheet_set_v_pagebreaks, {worksheet,breaks} )
end procedure

-- Set the order in which pages are printed.

global procedure worksheet_print_across( atom worksheet )
  c_proc( x_worksheet_print_across, {worksheet} )
end procedure

-- Set the option to display or hide gridlines on the screen and the printed page.

global procedure worksheet_gridlines( atom worksheet, atom option )
  c_proc( x_worksheet_gridlines, {worksheet,option} )
end procedure

-- Center the printed page horizontally.

global procedure worksheet_center_horizontally( atom worksheet )
  c_proc( x_worksheet_center_horizontally, {worksheet} )
end procedure

-- Center the printed page vertically.

global procedure worksheet_center_vertically( atom worksheet )
  c_proc( x_worksheet_center_vertically, {worksheet} )
end procedure

-- Set the option to print the row and column headers on the printed page.

global procedure worksheet_print_row_col_headers( atom worksheet )
  c_proc( x_worksheet_print_row_col_headers, {worksheet} )
end procedure

-- Set the number of rows to repeat at the top of each printed page.

global function worksheet_repeat_rows( atom worksheet, sequence rows )
  return c_func( x_worksheet_repeat_rows, {worksheet,rows[1],rows[2]} )
end function

-- Set the number of columns to repeat at the top of each printed page.

global function worksheet_repeat_columns( atom worksheet, sequence cols )
  return c_func( x_worksheet_repeat_columns, {worksheet,cols[1],cols[2]} )
end function

-- Set the print area for a worksheet.

global function worksheet_print_area( atom worksheet, sequence range )
  return c_func( x_worksheet_print_area, {worksheet,range[1],range[2],range[3],range[4]} )
end function

-- Fit the printed area to a specific number of pages both vertically and horizontally.

global procedure worksheet_fit_to_pages( atom worksheet, atom width, atom height )
  c_proc( x_worksheet_fit_to_pages, {worksheet,width,height} )
end procedure

-- Set the start page number when printing.

global procedure worksheet_set_start_page( atom worksheet, atom start_page )
  c_proc( x_worksheet_set_start_page, {worksheet,start_page} )
end procedure

-- Set the scale factor for the printed page.

global procedure worksheet_set_print_scale( atom worksheet, atom scale )
  c_proc( x_worksheet_set_print_scale, {worksheet,scale} )
end procedure

-- Set the font used in the cell.

global procedure format_set_font_name( atom format, object name )
  name = allocateString( name )
  c_proc( x_format_set_font_name, {format,name} )
  freeMemory(name)
end procedure

-- Set the size of the font used in the cell.

global procedure format_set_font_size( atom format, atom size )
  c_proc( x_format_set_font_size, {format,size} )
end procedure

-- Set the color of the font used in the cell.

global procedure format_set_font_color( atom format, atom color )
  c_proc( x_format_set_font_color, {format,color} )
end procedure

-- Turn on bold for the format font.

global procedure format_set_bold( atom format )
  c_proc( x_format_set_bold, {format} )
end procedure

-- Turn on italic for the format font.

global procedure format_set_italic( atom format )
  c_proc( x_format_set_italic, {format} )
end procedure

-- Turn on underline for the format.
--
-- The available underline styles are:
-- - LXW_UNDERLINE_SINGLE
-- - LXW_UNDERLINE_DOUBLE
-- - LXW_UNDERLINE_SINGLE_ACCOUNTING
-- - LXW_UNDERLINE_DOUBLE_ACCOUNTING

global procedure format_set_underline( atom format, atom style )
  c_proc( x_format_set_underline, {format,style} )
end procedure

-- Set the strikeout property of the font.

global procedure format_set_font_strikeout( atom format )
  c_proc( x_format_set_font_strikeout, {format} )
end procedure

-- Set the superscript/subscript property of the font.
--
-- The available script styles are:
-- - LXW_FONT_SUPERSCRIPT
-- - LXW_FONT_SUBSCRIPT

global procedure format_set_font_script( atom format, atom script )
  c_proc( x_format_set_font_script, {format,script} )
end procedure

-- Set the number format for a cell.

global procedure format_set_num_format( atom format, object num_format )
  num_format = allocateString( num_format )
  c_proc( x_format_set_num_format, {format,num_format} )
  freeMemory(num_format)
end procedure

-- Set the Excel built-in number format for a cell.
--
-- The Excel built-in number formats as shown in the table below:
--
--   | Index | Index | Format String                                        |
--   | ----- | ----- | ---------------------------------------------------- |
--   |   0   |  #00  | `General`                                            |
--   |   1   |  #01  | `0`                                                  |
--   |   2   |  #02  | `0.00`                                               |
--   |   3   |  #03  | `#,##0`                                              |
--   |   4   |  #04  | `#,##0.00`                                           |
--   |   5   |  #05  | `($#,##0_);($#,##0)`                                 |
--   |   6   |  #06  | `($#,##0_);[Red]($#,##0)`                            |
--   |   7   |  #07  | `($#,##0.00_);($#,##0.00)`                           |
--   |   8   |  #08  | `($#,##0.00_);[Red]($#,##0.00)`                      |
--   |   9   |  #09  | `0%`                                                 |
--   |  10   |  #0a  | `0.00%`                                              |
--   |  11   |  #0b  | `0.00E+00`                                           |
--   |  12   |  #0c  | `# ?/?`                                              |
--   |  13   |  #0d  | `# ??/??`                                            |
--   |  14   |  #0e  | `m/d/yy`                                             |
--   |  15   |  #0f  | `d-mmm-yy`                                           |
--   |  16   |  #10  | `d-mmm`                                              |
--   |  17   |  #11  | `mmm-yy`                                             |
--   |  18   |  #12  | `h:mm AM/PM`                                         |
--   |  19   |  #13  | `h:mm:ss AM/PM`                                      |
--   |  20   |  #14  | `h:mm`                                               |
--   |  21   |  #15  | `h:mm:ss`                                            |
--   |  22   |  #16  | `m/d/yy h:mm`                                        |
--   | ...   |  ...  | ...                                                  |
--   |  37   |  #25  | `(#,##0_);(#,##0)`                                   |
--   |  38   |  #26  | `(#,##0_);[Red](#,##0)`                              |
--   |  39   |  #27  | `(#,##0.00_);(#,##0.00)`                             |
--   |  40   |  #28  | `(#,##0.00_);[Red](#,##0.00)`                        |
--   |  41   |  #29  | `_(* #,##0_);_(* (#,##0);_(* "-"_);_(@_)`            |
--   |  42   |  #2a  | `_($* #,##0_);_($* (#,##0);_($* "-"_);_(@_)`         |
--   |  43   |  #2b  | `_(* #,##0.00_);_(* (#,##0.00);_(* "-"??_);_(@_)`    |
--   |  44   |  #2c  | `_($* #,##0.00_);_($* (#,##0.00);_($* "-"??_);_(@_)` |
--   |  45   |  #2d  | `mm:ss`                                              |
--   |  46   |  #2e  | `[h]:mm:ss`                                          |
--   |  47   |  #2f  | `mm:ss.0`                                            |
--   |  48   |  #30  | `##0.0E+0`                                           |
--   |  49   |  #31  | `@`                                                  |
--
-- Notes:
--  -  Numeric formats 23 to 36 are not documented by Microsoft and may differ in international
--     versions. The listed date and currency formats may also vary depending on system settings.
--
--  - The dollar sign in the above format appears as the defined local currency symbol.
--
--  - These formats can also be set via format_set_num_format().

global procedure format_set_num_format_index( atom format, atom index )
  c_proc( x_format_set_num_format_index, {format,index} )
end procedure

-- Set the cell unlocked state.

global procedure format_set_unlocked( atom format )
  c_proc( x_format_set_unlocked, {format} )
end procedure

-- Hide formulas in a cell.

global procedure format_set_hidden( atom format )
  c_proc( x_format_set_hidden, {format} )
end procedure


-- Set the alignment for data in the cell.
--
-- The following are the available horizontal alignments:
-- - LXW_ALIGN_LEFT
-- - LXW_ALIGN_CENTER
-- - LXW_ALIGN_RIGHT
-- - LXW_ALIGN_FILL
-- - LXW_ALIGN_JUSTIFY
-- - LXW_ALIGN_CENTER_ACROSS
-- - LXW_ALIGN_DISTRIBUTED
--
-- The following are the available vertical alignments:
-- - LXW_ALIGN_VERTICAL_TOP
-- - LXW_ALIGN_VERTICAL_BOTTOM
-- - LXW_ALIGN_VERTICAL_CENTER
-- - LXW_ALIGN_VERTICAL_JUSTIFY
-- - LXW_ALIGN_VERTICAL_DISTRIBUTED

global procedure format_set_align( atom format, atom align )
  c_proc( x_format_set_align, {format,align} )
end procedure

-- Wrap text in a cell.

global procedure format_set_text_wrap( atom format )
  c_proc( x_format_set_text_wrap, {format} )
end procedure

-- Set the rotation of the text in a cell.

global procedure format_set_rotation( atom format, atom angle )
  c_proc( x_format_set_rotation, {format,angle} )
end procedure

-- Set the cell text indentation level.

global procedure format_set_indent( atom format, atom indent )
  c_proc( x_format_set_indent, {format,indent} )
end procedure


-- Turn on the text "shrink to fit" for a cell.

global procedure format_set_shrink( atom format )
  c_proc( x_format_set_shrink, {format} )
end procedure

-- Set the background fill pattern for a cell
--
--   | Fill Type                     | Define                       |
--   | ----------------------------- | ---------------------------- |
--   | Solid                         | LXW_PATTERN_SOLID            |
--   | Medium gray                   | LXW_PATTERN_MEDIUM_GRAY      |
--   | Dark gray                     | LXW_PATTERN_DARK_GRAY        |
--   | Light gray                    | LXW_PATTERN_LIGHT_GRAY       |
--   | Dark horizontal line          | LXW_PATTERN_DARK_HORIZONTAL  |
--   | Dark vertical line            | LXW_PATTERN_DARK_VERTICAL    |
--   | Dark diagonal stripe          | LXW_PATTERN_DARK_DOWN        |
--   | Reverse dark diagonal stripe  | LXW_PATTERN_DARK_UP          |
--   | Dark grid                     | LXW_PATTERN_DARK_GRID        |
--   | Dark trellis                  | LXW_PATTERN_DARK_TRELLIS     |
--   | Light horizontal line         | LXW_PATTERN_LIGHT_HORIZONTAL |
--   | Light vertical line           | LXW_PATTERN_LIGHT_VERTICAL   |
--   | Light diagonal stripe         | LXW_PATTERN_LIGHT_DOWN       |
--   | Reverse light diagonal stripe | LXW_PATTERN_LIGHT_UP         |
--   | Light grid                    | LXW_PATTERN_LIGHT_GRID       | 
--   | Light trellis                 | LXW_PATTERN_LIGHT_TRELLIS    |
--   | 12.5% gray                    | LXW_PATTERN_GRAY_125         |
--   | 6.25% gray                    | LXW_PATTERN_GRAY_0625        |

global procedure format_set_pattern( atom format, atom index )
  c_proc( x_format_set_pattern, {format,index} )
end procedure

-- Set the pattern background color for a cell.

global procedure format_set_bg_color( atom format, atom color )
  c_proc( x_format_set_bg_color, {format,color} )
end procedure

-- Set the pattern foreground color for a cell.

global procedure format_set_fg_color( atom format, atom color )
  c_proc( x_format_set_fg_color, {format,color} )
end procedure

-- Set the cell border style.
--
-- The following border styles are available:
-- - LXW_BORDER_THIN
-- - LXW_BORDER_MEDIUM
-- - LXW_BORDER_DASHED
-- - LXW_BORDER_DOTTED
-- - LXW_BORDER_THICK
-- - LXW_BORDER_DOUBLE
-- - LXW_BORDER_HAIR
-- - LXW_BORDER_MEDIUM_DASHED
-- - LXW_BORDER_DASH_DOT
-- - LXW_BORDER_MEDIUM_DASH_DOT
-- - LXW_BORDER_DASH_DOT_DOT
-- - LXW_BORDER_MEDIUM_DASH_DOT_DOT
-- - LXW_BORDER_SLANT_DASH_DOT
--
-- The 'style' can be a single atom for all borders, or a sequence 
-- of four values for each border, e.g. {left, top, right, bottom}.

global procedure format_set_border( atom format, object style )
  if sequence( style ) then
    c_proc( x_format_set_left,  {format,style[1]} )
    c_proc( x_format_set_top,  {format,style[2]} )
    c_proc( x_format_set_right,  {format,style[3]} )
    c_proc( x_format_set_bottom, {format,style[4]} )
  else
    c_proc( x_format_set_border, {format,style} )
  end if
end procedure

-- Set the cell bottom border style.

global procedure format_set_bottom( atom format, atom style )
  c_proc( x_format_set_bottom, {format,style} )
end procedure

-- Set the cell top border style.

global procedure format_set_top( atom format, atom style )
  c_proc( x_format_set_top, {format,style} )
end procedure

-- Set the cell left border style.

global procedure format_set_left( atom format, atom style )
  c_proc( x_format_set_left, {format,style} )
end procedure

-- Set the cell right border style.

global procedure format_set_right( atom format, atom style )
  c_proc( x_format_set_right, {format,style} )
end procedure

-- Set the color of the cell border.
--
-- The 'color' can be a single atom for all borders, or a sequence 
-- of four values for each border, e.g. {left, top, right, bottom}.

global procedure format_set_border_color( atom format, object color )
  if sequence( color ) then
    c_proc( x_format_set_left_color,    {format,color[1]} )
    c_proc( x_format_set_top_color,    {format,color[2]} )
    c_proc( x_format_set_right_color,  {format,color[3]} )
    c_proc( x_format_set_bottom_color, {format,color[4]} )
  else
    c_proc( x_format_set_border_color, {format,color} )
  end if
end procedure

-- Set the color of the bottom cell border.

global procedure format_set_bottom_color( atom format, atom color )
  c_proc( x_format_set_bottom_color, {format,color} )
end procedure

-- Set the color of the top cell border.

global procedure format_set_top_color( atom format, atom color )
  c_proc( x_format_set_top_color, {format,color} )
end procedure

-- Set the color of the left cell border.

global procedure format_set_left_color( atom format, atom color )
  c_proc( x_format_set_left_color, {format,color} )
end procedure

-- Set the color of the right cell border.

global procedure format_set_right_color( atom format, atom color )
  c_proc( x_format_set_right_color, {format,color} )
end procedure

global procedure format_set_diag_type( atom format, atom value )
  c_proc( x_format_set_diag_type, {format,value} )
end procedure

global procedure format_set_diag_color( atom format, atom color )
  c_proc( x_format_set_diag_color, {format,color} )
end procedure

global procedure format_set_diag_border( atom format, atom value )
  c_proc( x_format_set_diag_border, {format,value} )
end procedure

global procedure format_set_font_outline( atom format )
  c_proc( x_format_set_font_outline, {format} )
end procedure

global procedure format_set_font_shadow( atom format )
  c_proc( x_format_set_font_shadow, {format} )
end procedure

global procedure format_set_font_family( atom format, atom value )
  c_proc( x_format_set_font_family, {format,value} )
end procedure

global procedure format_set_font_charset( atom format, atom value )
  c_proc( x_format_set_font_charset, {format,value} )
end procedure

global procedure format_set_font_scheme( atom format, object scheme )
  scheme = allocateString( scheme )
  c_proc( x_format_set_font_scheme, {format,scheme} )
  freeMemory(scheme)
end procedure

global procedure format_set_font_condense( atom format )
  c_proc( x_format_set_font_condense, {format} )
end procedure

global procedure format_set_font_extend( atom format )
  c_proc( x_format_set_font_extend, {format} )
end procedure

global procedure format_set_reading_order( atom format, atom value )
  c_proc( x_format_set_reading_order, {format,value} )
end procedure

global procedure format_set_theme( atom format, atom value )
  c_proc( x_format_set_theme, {format,value} )
end procedure

global function lxw_quote_sheetname( object str )
  atom ptr
  sequence result

  str = allocateString( str )
  ptr = c_func( x_lxw_quote_sheetname, {str} )
  result = peek_string( ptr )
  freeMemory(str)
  return result
end function

global function lxw_get_row( object row_str )
  object result

  row_str = allocateString( row_str )
  result = c_func( x_lxw_get_row, {row_str} )
  freeMemory(row_str)
  return result
end function

global function lxw_get_col( object row_str )
  object result

  row_str = allocateString( row_str )
  result = c_func( x_lxw_get_col, {row_str} )
  freeMemory(row_str)
  return result
end function

global function lxw_get_row_2( object row_str )
  object result

  row_str = allocateString( row_str )
  result = c_func( x_lxw_get_row_2, {row_str} )
  freeMemory(row_str)
  return result
end function

global function lxw_get_col_2( object row_str )
  object result

  row_str = allocateString( row_str )
  result = c_func( x_lxw_get_col_2, {row_str} )
  freeMemory(row_str)
  return result
end function

-- Convert an Excel `A1` cell string into a `(row, col)` sequence.

global function CELL( sequence cell )
  sequence result
  atom p_cell

  p_cell = allocateString( cell )
  result = { lxw_get_row(p_cell), lxw_get_col(p_cell) }
  freeMemory(p_cell)
  return result
end function

-- Convert an Excel `A:B` column range into a `(col1, col2)` sequence.

global function COLS( sequence cols )
  sequence result
  atom p_cols

  p_cols = allocateString( cols )
  result = { lxw_get_col(p_cols), lxw_get_col_2(p_cols) }
  freeMemory(p_cols)
  return result
end function

-- Convert an Excel `A1:B2` range into a `(first_row, first_col, last_row, last_col)` sequence.

global function RANGE( sequence range )
  sequence result
  atom p_range

  p_range = allocateString( range )
  result = { lxw_get_row(range), lxw_get_col(range), lxw_get_row_2(range), lxw_get_col_2(range) }
  freeMemory(p_range)
  return result
end function

-- Convert an Excel `A1:B2` range into a `(row1, row2)` sequence.

global function ROWS( sequence rows )
  sequence result
  atom p_rows

  p_rows = allocateString( rows )
  result = { lxw_get_row(p_rows), lxw_get_row_2(p_rows) }
  freeMemory(p_rows)
  return result
end function
